MAINNET_NETWORK_ID = 1029
TESTNET_NETWORK_ID = 1
TYPE = "type"
MAINNET_PREFIX = "cfx"
TESTNET_PREFIX = "cfxtest"
COMMON_NET_PREFIX = "net"

TYPE_NULL = "null"
TYPE_BUILTIN = "builtin"
TYPE_USER = "user"
TYPE_CONTRACT = "contract"
TYPE_INVALID = "invalid"

HEX_PREFIX = "0x"
DELIMITER = ":"

VERSION_BYTE = bytes(1)
CHECKSUM_TEMPLATE = bytes(8)
